/**
 * @ingroup    uCTest
 * @defgroup   macro Macro: Macro helper definitions
 *
 * @brief macro helper definitions
 *
 * @author     Sebastian Koschmieder <sep@seplog.org>
 * @date       2021-01-21 16:29:18
 * @{
 */

#include <avr/io.h>
#include <inttypes.h>
#include <stdlib.h>

#ifdef __cplusplus
extern "C" {
#endif

#ifndef _UCTEST_UCTEST_MACRO_H_
#if !defined( __DOXYGEN__ )
#define _UCTEST_UCTEST_MACRO_H_
#endif

#if !defined( __DOXYGEN__ )
#define _UCTEST_STR_IMPL( x ) #x
#define UCTEST_STR( x ) _UCTEST_STR_IMPL( x )
#endif

#if defined( __DOXYGEN__ )
/**
 * @brief Place a string constant in an optimal storage for a microcontroller
 *
 * This should place a string constant in an optimal storage for the microcontroller. for the AVR
 * microcontroller this is the flash.
 *
 * @param[in] X The string constant
 */
#define UCTEST_XSTR( X )
#else
#define UCTEST_XSTR( X )                     \
    ( {                                      \
        static const __flash char __s[] = X; \
        &__s[ 0 ];                           \
    } )
#endif

#if defined( __DOXYGEN__ )
/**
 * @brief Calls a function foreach parameter in the variadic argument list
 *
 * This calls a function for each parameter in the variadic argument list. Every single argument
 * from the variadic argument list is used as parameter for the function to call given in the first
 * parameter.
 *
 * @param[in] x Macro or function to call
 * @param[in] ... Arguments for the macro or function
 */
#define UCTEST_FOREACH( x, ... )
#else
#define UCTEST_FOREACH_0( X )
#define UCTEST_FOREACH_1( WHAT, X, ... ) WHAT( X )
#define UCTEST_FOREACH_2( WHAT, X, ... ) \
    WHAT( X );                           \
    UCTEST_FOREACH_1( WHAT, __VA_ARGS__ )
#define UCTEST_FOREACH_3( WHAT, X, ... ) \
    WHAT( X );                           \
    UCTEST_FOREACH_2( WHAT, __VA_ARGS__ )

#define UCTEST_FOREACH_4( WHAT, X, ... ) \
    WHAT( X );                           \
    UCTEST_FOREACH_3( WHAT, __VA_ARGS__ )
#define UCTEST_FOREACH_5( WHAT, X, ... ) \
    WHAT( X );                           \
    UCTEST_FOREACH_4( WHAT, __VA_ARGS__ )
#define UCTEST_FOREACH_6( WHAT, X, ... ) \
    WHAT( X );                           \
    UCTEST_FOREACH_5( WHAT, __VA_ARGS__ )
#define UCTEST_FOREACH_7( WHAT, X, ... ) \
    WHAT( X );                           \
    UCTEST_FOREACH_6( WHAT, __VA_ARGS__ )
#define UCTEST_FOREACH_8( WHAT, X, ... ) \
    WHAT( X );                           \
    UCTEST_FOREACH_7( WHAT, __VA_ARGS__ )
#define UCTEST_FOREACH_9( WHAT, X, ... ) \
    WHAT( X );                           \
    UCTEST_FOREACH_8( WHAT, __VA_ARGS__ )
#define UCTEST_FOREACH_10( WHAT, X, ... ) \
    WHAT( X );                            \
    UCTEST_FOREACH_9( WHAT, __VA_ARGS__ )
#define UCTEST_FOREACH_11( WHAT, X, ... ) \
    WHAT( X );                            \
    UCTEST_FOREACH_10( WHAT, __VA_ARGS__ )
#define UCTEST_FOREACH_12( WHAT, X, ... ) \
    WHAT( X );                            \
    UCTEST_FOREACH_11( WHAT, __VA_ARGS__ )

#define UCTEST_GET_NTH_ARG( _11, _10, _9, _8, _7, _6, _5, _4, _3, _2, _1, _0, N, ... ) N

#define UCTEST_FOREACH( x, ... )                                                                   \
    UCTEST_GET_NTH_ARG( "", ##__VA_ARGS__, UCTEST_FOREACH_11, UCTEST_FOREACH_10, UCTEST_FOREACH_9, \
                        UCTEST_FOREACH_8, UCTEST_FOREACH_7, UCTEST_FOREACH_6, UCTEST_FOREACH_5,    \
                        UCTEST_FOREACH_4, UCTEST_FOREACH_3, UCTEST_FOREACH_2, UCTEST_FOREACH_1,    \
                        UCTEST_FOREACH_0 )                                                         \
    ( x, ##__VA_ARGS__ )
#endif

#if defined( __DOXYGEN__ )
/**
 * @brief Check if the function name is present at compile time
 *
 * Check if the function name is present at compile time
 *
 * @param[in] name The name of the function to check
 */
#define UCTEST_CHECK_FUNCTION_PRESENT( name )
#else
#define UCTEST_CHECK_FUNCTION_PRESENT( name )      \
    do {                                           \
        extern void name( void );                  \
        asm( "; check presence" : : "r"( name ) ); \
    } while( 0 )
#endif

#if defined( __DOXYGEN__ )
/**
 * @brief Decide what to call in respect of the argument number
 *
 * Calls default if there are no variadic arguments present. Otherwise calls args
 *
 * @param[in] default Call if no args present
 * @param[in] args called if minimal one arg is present
 * @param ... the arguments
 */
#define UCTEST_CALL_DEFARGS( default, args, ... )
#else
#define UCTEST_CALL_DEFARGS( default, args, ... )                                                \
    UCTEST_GET_NTH_ARG( "", ##__VA_ARGS__, args, args, args, args, args, args, args, args, args, \
                        args, args, default )
#endif

#if defined( __DOXYGEN__ )
/**
 * @brief Concat two names
 *
 * @param[in] x the first name
 * @param[in] y the second name
 */
#define USTEST_CONCAT( x, y )
#else
#define _UCTEST_CONCAT_IMPL( x, y ) x##y
#define UCTEST_CONCAT( x, y ) _UCTEST_CONCAT_IMPL( x, y )
#endif

#endif /** _UCTEST_UCTEST_MACRO_H_ */

#ifdef __cplusplus
}
#endif
/** @} */
